@@ -7,13 +7,73 @@ use anyhow::{anyhow, Context, Error};
 use once_cell::sync::Lazy;
 use parking_lot::Mutex;
 use resolve::PluginCache;
-use swc_common::{collections::AHashMap, plugin::Serialized};
-use wasmer::{imports, Array, Exports, Instance, Memory, MemoryType, Module, Store, WasmPtr};
+use swc_common::{
+    collections::AHashMap,
+    errors::{Diagnostic, HANDLER},
+    plugin::{PluginError, Serialized},
+};
+use wasmer::{
+    imports, Array, Exports, Function, Instance, LazyInit, Memory, Module, Store, WasmPtr,
+};
 use wasmer_cache::{Cache, Hash};
-use wasmer_wasi::{get_wasi_version, WasiState};
+use wasmer_wasi::{is_wasi_module, WasiState};
 
 pub mod resolve;
 
+fn copy_bytes_into_host(memory: &Memory, bytes_ptr: i32, bytes_ptr_len: i32) -> Vec<u8> {
+    let ptr: WasmPtr<u8, Array> = WasmPtr::new(bytes_ptr as _);
+
+    // Deref & read through plugin's wasm memory space via returned ptr
+    let derefed_ptr = ptr
+        .deref(memory, 0, bytes_ptr_len as u32)
+        .expect("Should able to deref from given ptr");
+
+    derefed_ptr
+        .iter()
+        .enumerate()
+        .take(bytes_ptr_len as usize)
+        .map(|(_size, cell)| cell.get())
+        .collect::<Vec<u8>>()
+}
+
+/// Set plugin's transformed result into host's enviroment.
+/// This is an `imported` fn - when we instantiate plugin module, we inject this
+/// fn into pluging's export space. Once transform completes, plugin will call
+/// this to set its result back to host.
+fn set_transform_result(env: &HostEnvironment, bytes_ptr: i32, bytes_ptr_len: i32) {
+    if let Some(memory) = env.memory_ref() {
+        (*env.transform_result.lock()) = copy_bytes_into_host(memory, bytes_ptr, bytes_ptr_len);
+    }
+}
+
+fn emit_diagnostics(env: &HostEnvironment, bytes_ptr: i32, bytes_ptr_len: i32) {
+    if let Some(memory) = env.memory_ref() {
+        if HANDLER.is_set() {
+            HANDLER.with(|handler| {
+                let diagnostics_bytes = copy_bytes_into_host(memory, bytes_ptr, bytes_ptr_len);
+                let serialized = Serialized::new_for_plugin(&diagnostics_bytes[..], bytes_ptr_len);
+                let diagnostic = Serialized::deserialize::<Diagnostic>(&serialized)
+                    .expect("Should able to be deserialized into diagnsotic");
+
+                let mut builder =
+                    swc_common::errors::DiagnosticBuilder::new_diagnostic(handler, diagnostic);
+                builder.emit();
+            })
+        }
+    }
+}
+
+#[derive(wasmer::WasmerEnv, Clone)]
+/// An external enviornment state imported (declared in host, injected into
+/// guest) fn can access. This'll allow host access updated state via plugin's
+/// transform.
+/// ref: https://docs.wasmer.io/integrations/examples/host-functions#declaring-the-data
+struct HostEnvironment {
+    #[wasmer(export)]
+    memory: wasmer::LazyInit<Memory>,
+    transform_result: Arc<Mutex<Vec<u8>>>,
+}
+
 /// Load plugin from specified path.
 /// If cache is provided, it'll try to load from cache first to avoid
 /// compilation.
@@ -27,7 +87,10 @@ pub mod resolve;
 /// [This code](https://github.com/swc-project/swc/blob/fc4c6708f24cda39640fbbfe56123f2f6eeb2474/crates/swc/src/plugin.rs#L19-L44)
 /// includes previous incorrect attempt to workaround file read issues.
 /// In actual transform, `plugins` is also being called per each transform.
-fn load_plugin(plugin_path: &Path, cache: &mut Option<PluginCache>) -> Result<Instance, Error> {
+fn load_plugin(
+    plugin_path: &Path,
+    cache: &mut Option<PluginCache>,
+) -> Result<(Instance, Arc<Mutex<Vec<u8>>>), Error> {
     static BYTE_CACHE: Lazy<Mutex<AHashMap<PathBuf, Arc<Vec<u8>>>>> = Lazy::new(Default::default);
 
     // TODO: This caching streategy does not consider few edge cases.
@@ -96,11 +159,27 @@ fn load_plugin(plugin_path: &Path, cache: &mut Option<PluginCache>) -> Result<In
 
     return match module {
         Ok(module) => {
-            let memory = Memory::new(&wasmer_store, MemoryType::new(1, None, false))?;
-            let wasi_version = get_wasi_version(&module, true);
+            let transform_result: Arc<Mutex<Vec<u8>>> = Arc::new(Mutex::new(vec![]));
+            let set_transform_result_fn_decl = Function::new_native_with_env(
+                &wasmer_store,
+                HostEnvironment {
+                    memory: LazyInit::default(),
+                    transform_result: transform_result.clone(),
+                },
+                set_transform_result,
+            );
 
-            // Assume plugin binary targets wasm32-wasi if wasi version can be found
-            let import_object = if wasi_version.is_some() {
+            let emit_diagnostics_fn_decl = Function::new_native_with_env(
+                &wasmer_store,
+                HostEnvironment {
+                    memory: LazyInit::default(),
+                    transform_result: transform_result.clone(),
+                },
+                emit_diagnostics,
+            );
+
+            // Plugin binary can be either wasm32-wasi or wasm32-unknown-unknown
+            let import_object = if is_wasi_module(&module) {
                 // Create the `WasiEnv`.
                 let mut wasi_env = WasiState::new(
                     plugin_path
@@ -113,10 +192,11 @@ fn load_plugin(plugin_path: &Path, cache: &mut Option<PluginCache>) -> Result<In
                 // Generate an `ImportObject` from wasi_env
                 let mut import_object = wasi_env.import_object(&module)?;
 
-                // Inject `memory` into generated import_object, this is required to allocate
-                // guest (plugin)'s memory space from the host (SWC).
+                // This'll inject few interfaces into plugin's namespace to let plugin
+                // communicate with host
                 let mut env = Exports::new();
-                env.insert("memory", memory);
+                env.insert("__set_transform_result", set_transform_result_fn_decl);
+                env.insert("__emit_diagnostics", emit_diagnostics_fn_decl);
                 import_object.register("env", env);
                 import_object
             }
@@ -125,12 +205,15 @@ fn load_plugin(plugin_path: &Path, cache: &mut Option<PluginCache>) -> Result<In
             else {
                 imports! {
                     "env" => {
-                        "memory" => memory
+                        "__set_transform_result" => set_transform_result_fn_decl,
+                        "__emit_diagnostics" => emit_diagnostics_fn_decl
                     }
                 }
             };
 
-            Instance::new(&module, &import_object).context("Failed to create plugin instance")
+            Instance::new(&module, &import_object)
+                .map(|i| (i, transform_result))
+                .context("Failed to create plugin instance")
         }
         Err(err) => Err(err),
     };
@@ -140,7 +223,7 @@ fn load_plugin(plugin_path: &Path, cache: &mut Option<PluginCache>) -> Result<In
 /// teardown
 struct PluginTransformTracker {
     // Main transform interface plugin exports
-    exported_plugin_transform: wasmer::NativeFunc<(i32, i32, i32, i32), (i32, i32)>,
+    exported_plugin_transform: wasmer::NativeFunc<(i32, i32, i32, i32), i32>,
     // `__free` function automatically exported via swc_plugin sdk to allow deallocation in guest
     // memory space
     exported_plugin_free: wasmer::NativeFunc<(i32, i32), i32>,
@@ -150,16 +233,17 @@ struct PluginTransformTracker {
     instance: Instance,
     // Reference to the pointers succesfully allocated which'll be freed by Drop.
     allocated_ptr_vec: Vec<(i32, i32)>,
+    transform_result: Arc<Mutex<Vec<u8>>>,
 }
 
 impl PluginTransformTracker {
     fn new(path: &Path, cache: &mut Option<PluginCache>) -> Result<PluginTransformTracker, Error> {
-        let instance = load_plugin(path, cache)?;
+        let (instance, transform_result) = load_plugin(path, cache)?;
 
         let tracker = PluginTransformTracker {
             exported_plugin_transform: instance
                 .exports
-                .get_native_function::<(i32, i32, i32, i32), (i32, i32)>("__plugin_process_impl")?,
+                .get_native_function::<(i32, i32, i32, i32), i32>("__plugin_process_impl")?,
             exported_plugin_free: instance
                 .exports
                 .get_native_function::<(i32, i32), i32>("__free")?,
@@ -168,6 +252,7 @@ impl PluginTransformTracker {
                 .get_native_function::<u32, i32>("__alloc")?,
             instance,
             allocated_ptr_vec: Vec::with_capacity(3),
+            transform_result,
         };
 
         Ok(tracker)
@@ -209,23 +294,28 @@ impl PluginTransformTracker {
 
     /// Copy guest's memory into host, construct serialized struct from raw
     /// bytes.
-    fn read_bytes_from_guest(&mut self, returned_ptr: i32, len: i32) -> Result<Serialized, Error> {
-        let memory = self.instance.exports.get_memory("memory")?;
-        let ptr: WasmPtr<u8, Array> = WasmPtr::new(returned_ptr as _);
+    fn read_bytes_from_guest(&mut self, returned_ptr_result: i32) -> Result<Serialized, Error> {
+        let transformed_result = &(*self.transform_result.lock());
+        let ret =
+            Serialized::new_for_plugin(&transformed_result[..], transformed_result.len() as i32);
 
-        // Deref & read through plugin's wasm memory space via returned ptr
-        let derefed_ptr = ptr
-            .deref(memory, 0, len.try_into()?)
-            .ok_or(anyhow!("Failed to deref raw bytes from plugin's memory"))?;
-
-        let transformed_raw_bytes = derefed_ptr
-            .iter()
-            .enumerate()
-            .take(len.try_into()?)
-            .map(|(_size, cell)| cell.get())
-            .collect::<Vec<u8>>();
-
-        Ok(Serialized::new_for_plugin(&transformed_raw_bytes[..], len))
+        if returned_ptr_result == 0 {
+            Ok(ret)
+        } else {
+            let err: PluginError = Serialized::deserialize(&ret)?;
+            match err {
+                PluginError::SizeInteropFailure(msg) => Err(anyhow!(
+                    "Failed to convert pointer size to calculate: {}",
+                    msg
+                )),
+                PluginError::Deserialize((msg, ..)) | PluginError::Serialize(msg) => {
+                    Err(anyhow!("{}", msg))
+                }
+                _ => Err(anyhow!(
+                    "Unexpected error occurred while running plugin transform"
+                )),
+            }
+        }
     }
 
     fn transform(
@@ -236,15 +326,14 @@ impl PluginTransformTracker {
         let guest_program_ptr = self.write_bytes_into_guest(program)?;
         let config_str_ptr = self.write_bytes_into_guest(config)?;
 
-        let returned_ptr = self.exported_plugin_transform.call(
+        let result = self.exported_plugin_transform.call(
             guest_program_ptr.0,
             guest_program_ptr.1,
             config_str_ptr.0,
             config_str_ptr.1,
         )?;
 
-        self.allocated_ptr_vec.push(returned_ptr);
-        self.read_bytes_from_guest(returned_ptr.0, returned_ptr.1)
+        self.read_bytes_from_guest(result)
     }
 }
 
@@ -258,8 +347,6 @@ impl Drop for PluginTransformTracker {
     }
 }
 
-// TODO
-// - error propagation from plugin
 pub fn apply_js_plugin(
     plugin_name: &str,
     path: &Path,
